# Copyright 2009 Paul Hummer
# Copyright 2009 Canonical Ltd.
#
# This file is part of Tarmac.
#
# Authors: Paul Hummer
#          Rodney Dawes
#
# Tarmac is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by
# the Free Software Foundation.
#
# Tarmac is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tarmac.  If not, see <http://www.gnu.org/licenses/>.

'''Tarmac branch tools.'''
import logging
import os
import shutil
import tempfile

from bzrlib import branch as bzr_branch
from bzrlib.errors import NoSuchRevision
from bzrlib.workingtree import WorkingTree

from tarmac.config import BranchConfig
from tarmac.exceptions import BranchHasConflicts, TarmacMergeError


class Branch(object):

    def __init__(self, lp_branch, config=False, target=None):
        self.lp_branch = lp_branch
        if config:
            self.config = BranchConfig(lp_branch.bzr_identity, config)
        else:
            self.config = None

        self.target = target
        self.logger = logging.getLogger('tarmac')
        self.temp_tree_dir = None

    def get_bzr_branch(self):
        return bzr_branch.Branch.open(self.lp_branch.bzr_identity)

    # For backwards compatibility
    bzr_branch = property(get_bzr_branch)

    def get_tree(self):
        if self.temp_tree_dir is not None:
            return WorkingTree.open(self.temp_tree_dir)            
        if os.path.exists(self.config.tree_dir):
            return WorkingTree.open(self.config.tree_dir)

    # For backwards compatibility
    tree = property(get_tree)

    def __del__(self):
        """Do some potenetially necessary cleanup during deletion."""
        if self.temp_tree_dir is not None:
            shutil.rmtree(self.temp_tree_dir)

    @classmethod
    def create(cls, lp_branch, config, create_tree=False, target=None):
        clazz = cls(lp_branch, config, target)
        if create_tree:
            clazz.create_tree()
        return clazz


    def create_tree(self):
        '''Create the dir and working tree.'''
        bzr_branch = self.get_bzr_branch()
        try:
            tree = self.get_tree()
            if tree is None:
                self.logger.debug('Tree does not exist.  Creating dir')
                # Create the path up to but not including tree_dir if it does
                # not exist.
                parent_dir = os.path.dirname(self.config.tree_dir)
                if not os.path.exists(parent_dir):
                    os.makedirs(parent_dir)
                tree = bzr_branch.create_checkout(
                    self.config.tree_dir, lightweight=False)
            self.logger.debug(
                'Using tree in %(tree_dir)s' % {
                    'tree_dir': self.config.tree_dir})
        except AttributeError:
            # Store this so we can rmtree later
            self.temp_tree_dir = tempfile.mkdtemp()
            self.logger.debug(
                'Using temp dir at %(tree_dir)s' % {
                    'tree_dir': self.temp_tree_dir})
            tree = bzr_branch.create_checkout(self.temp_tree_dir)

        self.cleanup()

    def cleanup(self):
        '''Remove the working tree from the temp dir.'''
        tree = self.get_tree()
        assert tree
        self.logger.info("Running cleanup in %s." % (
            self.lp_branch.bzr_identity))
        tree.revert()
        self.logger.info("Reverted changes in %s." % (
            self.lp_branch.bzr_identity))
        for filename in [tree.abspath(f) for f in self.unmanaged_files]:
            if os.path.isdir(filename) and not os.path.islink(filename):
                shutil.rmtree(filename)
            else:
                os.remove(filename)

        self.logger.info("Successfully removed extra files from %s." % (
            self.lp_branch.bzr_identity))
        tree.update()

    def merge(self, branch, revid=None):
        '''Merge from another tarmac.branch.Branch instance.'''
        tree = self.get_tree()
        assert tree
        conflict_list = tree.merge_from_branch(
            branch.get_bzr_branch(), to_revision=revid)
        if conflict_list:
            message = u'Conflicts merging branch.'
            lp_comment = (
                u'Attempt to merge into %(target)s failed due to conflicts: '
                u'\n\n%(output)s' % {
                    'target': self.lp_branch.display_name,
                    "output": self.conflicts})
            raise BranchHasConflicts(message, lp_comment)

    @property
    def unmanaged_files(self):
        """Get the list of ignored and unknown files in the tree."""
        tree = self.get_tree()
        assert tree
        tree.lock_read()
        unmanaged = [x for x in tree.unknowns()]
        unmanaged.extend([x[0] for x in tree.ignored_files()])
        tree.unlock()
        return unmanaged

    @property
    def conflicts(self):
        '''Print the conflicts.'''
        tree = self.get_tree()
        assert tree
        conflicts = []
        for conflict in tree.conflicts():
            conflicts.append(
                u'%s in %s' % (conflict.typestring, conflict.path))
        return '\n'.join(conflicts)

    def commit(self, commit_message, revprops=None, **kwargs):
        '''Commit changes.'''
        tree = self.get_tree()
        assert tree
        if not revprops:
            revprops = {}

        authors = kwargs.pop('authors', None)
        reviews = kwargs.pop('reviews', None)

        if not authors:
            authors = self.authors

        if reviews:
            for review in reviews:
                if '\n' in review:
                    raise TarmacMergeError('\\n is not a valid character in a '
                                           'review identity or vote.')
            revprops['reviews'] = '\n'.join(reviews)

        tree.commit(commit_message, committer='Tarmac',
                    revprops=revprops, authors=authors)

    @property
    def landing_candidates(self):
        '''Wrap the LP representation of landing_candidates.'''
        return self.lp_branch.landing_candidates

    @property
    def authors(self):
        author_list = []

        bzr_branch = self.get_bzr_branch()
        if self.target:
            target_bzr_branch = self.target.get_bzr_branch()
            bzr_branch.lock_read()
            target_bzr_branch.lock_read()

            graph = bzr_branch.repository.get_graph(
                target_bzr_branch.repository)

            unique_ids = graph.find_unique_ancestors(
                bzr_branch.last_revision(),
                [target_bzr_branch.last_revision()])

            revs = bzr_branch.repository.get_revisions(unique_ids)
            for rev in revs:
                apparent_authors = rev.get_apparent_authors()
                for author in apparent_authors:
                    author.replace('\n', '')
                    if author not in author_list:
                        author_list.append(author)

            target_bzr_branch.unlock()
            bzr_branch.unlock()
        else:
            last_rev = bzr_branch.last_revision()
            if last_rev != 'null:':
                rev = bzr_branch.repository.get_revision(last_rev)
                apparent_authors = rev.get_apparent_authors()
                author_list.extend(
                    [a.replace('\n', '') for a in apparent_authors])

        return author_list

    @property
    def fixed_bugs(self):
        """Return the list of bugs fixed by the branch."""
        bugs_list = []

        bzr_branch = self.get_bzr_branch()
        bzr_branch.lock_read()
        oldrevid = bzr_branch.get_rev_id(self.lp_branch.revision_count)
        for rev_info in bzr_branch.iter_merge_sorted_revisions(
            stop_revision_id=oldrevid):
            try:
                rev = bzr_branch.repository.get_revision(rev_info[0])
                for bug in rev.iter_bugs():
                    if bug[0].startswith('https://launchpad.net/bugs/'):
                        bugs_list.append(bug[0].replace(
                                'https://launchpad.net/bugs/', ''))
            except NoSuchRevision:
                continue

        bzr_branch.unlock()
        return bugs_list
